@@ -229,29 +229,21 @@
 #+sb-assembling
 (define-assembly-routine (call-symbol
                           (:return-style :none))
-    ((:temp fun (any-reg descriptor-reg) rax-offset)
-     (:temp length (any-reg descriptor-reg) rax-offset)
-     (:temp vector (any-reg descriptor-reg) rbx-offset))
-  (%lea-for-lowtag-test vector fun other-pointer-lowtag)
-  (inst test :byte vector lowtag-mask)
+    ((:temp fun (any-reg descriptor-reg) rax-offset) ; FUN = the symbol
+     (:temp fdefn (any-reg descriptor-reg) rbx-offset))
+  (%lea-for-lowtag-test fdefn fun other-pointer-lowtag)
+  (inst test :byte fdefn lowtag-mask)
   (inst jmp :nz not-callable)
   (inst cmp :byte (ea (- other-pointer-lowtag) fun) symbol-widetag)
   (inst jmp :ne not-callable)
-  (load-symbol-info-vector vector fun)
-  ;; info-vector-fdefn
-  (inst cmp vector nil-value)
-  (inst jmp :e undefined)
 
-  (inst mov :dword r10-tn (ea (- (* 2 n-word-bytes) other-pointer-lowtag) vector))
-  (inst and :dword r10-tn (fixnumize (1- (ash 1 (* info-number-bits 2)))))
-  (inst cmp :dword r10-tn (fixnumize (1+ (ash +fdefn-info-num+ info-number-bits))))
-  (inst jmp :b undefined)
-
-  (inst mov vector-len-op-size length (vector-len-ea vector))
-  (inst mov fun (ea (- 8 other-pointer-lowtag) vector length
-                    (ash 1 (- word-shift n-fixnum-tag-bits))))
-
-  (inst jmp (ea (- (* fdefn-raw-addr-slot n-word-bytes) other-pointer-lowtag) fun))
+  (loadw fdefn fun symbol-fdefn-slot other-pointer-lowtag)
+  (inst test :dword fdefn fdefn)
+  (inst jmp :z UNDEFINED)
+  ;; I think we need this MOV because the undefined-function trap examines RAX
+  ;; to see what FDEFN you tried to call through.
+  (inst mov fun fdefn)
+  (inst jmp (ea (- (* fdefn-raw-addr-slot n-word-bytes) other-pointer-lowtag) fdefn))
   UNDEFINED
   (inst jmp (make-fixup 'undefined-tramp :assembly-routine))
   NOT-CALLABLE
@@ -268,9 +260,10 @@
                          ((:arg target (descriptor-reg any-reg) rdx-offset)
                           (:arg start any-reg rbx-offset)
                           (:arg count any-reg rcx-offset)
+                          (:temp bsp-temp any-reg r11-offset)
                           (:temp catch any-reg rax-offset))
 
-  (declare (ignore start count))
+  (declare (ignore start count bsp-temp))
 
   (load-tl-symbol-value catch *current-catch-block*)
 
@@ -323,6 +316,7 @@
                           (:temp where unsigned-reg r8-offset)
                           (:temp symbol unsigned-reg r9-offset)
                           (:temp value unsigned-reg r10-offset)
+                          (:temp bsp-temp unsigned-reg r11-offset)
                           (:temp zero complex-double-reg float0-offset))
   AGAIN
   (let ((error (generate-error-code nil 'invalid-unwind-error)))
@@ -349,7 +343,7 @@
   ;; run it twice) one of the variables being unbound can be
   ;; *interrupts-enabled*
   (loadw where uwp unwind-block-bsp-slot)
-  (unbind-to-here where symbol value temp-reg-tn zero)
+  (unbind-to-here where symbol value bsp-temp zero)
 
   ;; Set next unwind protect context.
   (loadw block uwp unwind-block-uwp-slot)
@@ -370,7 +364,7 @@
 
   DO-EXIT
   (loadw where block unwind-block-bsp-slot)
-  (unbind-to-here where symbol value temp-reg-tn zero)
+  (unbind-to-here where symbol value bsp-temp zero)
 
   (loadw rbp-tn block unwind-block-cfp-slot)
 
@@ -388,99 +382,91 @@
     ;; to stack, because we do do not give the register allocator access to it.
     ;; And call_into_lisp saves it as per convention, not that it matters,
     ;; because there's no way to get back into C code anyhow.
-    (inst mov thread-base-tn (ea (make-fixup "all_threads" :foreign-dataref)))
-    (inst mov thread-base-tn (ea thread-base-tn))))
+    (inst mov thread-tn (ea (make-fixup "all_threads" :foreign-dataref)))
+    (inst mov thread-tn (ea thread-tn))))
 
-;;; Perform a store to code, updating the GC page (card) protection bits.
-;;; This is not a "good" implementation of soft card marking.
-;;; It is used *only* for pages of code. The real implementation (work in
-;;; progress) will differ in at least these ways:
-;;; - there will be no use of pseudo-atomic
-;;; - stores will be inlined without the helper routine below
-;;; - will be insensitive to the size of a page table entry
-;;; - will avoid use of a :lock prefix by allocating 1 byte per mark
-;;; - won't need to subtract the heap base or compare to the card count
-;;;   to compute the mark address, so will use fewer instructions.
-;;; It is similar in that for code objects (indeed most objects
-;;; except simple-vectors), it marks the object header which is
-;;; not always on the same GC card affected by the store operation
-;;;
+;;; Perform a store to code, updating the GC card mark bit.
+;;; This has two additional complications beyond the ordinary
+;;; generational barrier:
+;;; 1. immobile code uses its own card table which maps linearly
+;;;    with the page index, unlike the dynamic space card table
+;;;    that has a different way of computing a card address.
+;;; 2. code objects are so seldom written that it behooves us to
+;;;    track within each object whether it has been written,
+;;;    thereby avoiding scanning of unwritten objects.
+;;;    This is especially important for immobile space where
+;;;    it is likely that new code will be co-located on a page
+;;;    with old code due to the non-moving allocator.
 #+sb-assembling
 (define-assembly-routine (code-header-set (:return-style :none)) ()
-  (inst push rax-tn)
-  (inst push rdi-tn)
-  ;; stack: spill[2], ret-pc, object, index, value-to-store
-
+  ;; stack: ret-pc, object, index, value-to-store
+  (symbol-macrolet ((object (ea 8 rsp-tn))
+                    (word-index (ea 16 rsp-tn))
+                    (newval (ea 24 rsp-tn))
+                    ;; these are declared as vop temporaries
+                    (rax rax-tn)
+                    (rdx rdx-tn)
+                    (rdi rdi-tn))
   (ensure-thread-base-tn-loaded)
   (pseudo-atomic ()
     (assemble ()
       #+immobile-space
       (progn
-        (inst mov temp-reg-tn (ea 24 rsp-tn))
-        (inst sub temp-reg-tn (thread-slot-ea thread-varyobj-space-addr-slot))
-        (inst shr temp-reg-tn (1- (integer-length immobile-card-bytes)))
-        (inst cmp temp-reg-tn (thread-slot-ea thread-varyobj-card-count-slot))
+        (inst mov rax object)
+        (inst sub rax (thread-slot-ea thread-varyobj-space-addr-slot))
+        (inst shr rax (1- (integer-length immobile-card-bytes)))
+        (inst cmp rax (thread-slot-ea thread-varyobj-card-count-slot))
         (inst jmp :ae try-dynamic-space)
-        (inst mov rdi-tn (thread-slot-ea thread-varyobj-card-marks-slot))
-        (inst bts :dword :lock (ea rdi-tn) temp-reg-tn)
+        (inst mov rdi (thread-slot-ea thread-varyobj-card-marks-slot))
+        (inst bts :dword :lock (ea rdi-tn) rax)
         (inst jmp store))
-
       TRY-DYNAMIC-SPACE
-      (inst mov temp-reg-tn (ea 24 rsp-tn)) ; reload
-      (inst sub temp-reg-tn (thread-slot-ea thread-dynspace-addr-slot))
-      (inst shr temp-reg-tn (1- (integer-length gencgc-card-bytes)))
-      (inst cmp temp-reg-tn (thread-slot-ea thread-dynspace-card-count-slot))
-      (inst jmp :ae store) ; neither dynamic nor immobile space. (weird!)
-
-      ;; sizeof (struct page) depends on GENCGC-CARD-BYTES
-      ;; It's 4+2+1+1 = 8 bytes if GENCGC-CARD-BYTES is (unsigned-byte 16),
-      ;; or   4+4+1+1 = 10 bytes (rounded to 12) if wider than (unsigned-byte 16).
-      ;; See the corresponding alien structure definition in 'room.lisp'
-      (cond ((typep gencgc-card-bytes '(unsigned-byte 16))
-             (inst shl temp-reg-tn 3) ; multiply by 8
-             (inst add temp-reg-tn (thread-slot-ea thread-dynspace-pte-base-slot))
-             ;; clear WP - bit index 5 of flags byte
-             (inst and :byte :lock (ea 6 temp-reg-tn) (lognot (ash 1 5))))
-            (t
-             (inst lea temp-reg-tn (ea temp-reg-tn temp-reg-tn 2)) ; multiply by 3
-             (inst shl temp-reg-tn 2) ; then by 4, = 12
-             (inst add temp-reg-tn (thread-slot-ea thread-dynspace-pte-base-slot))
-             ;; clear WP
-             (inst and :byte :lock (ea 8 temp-reg-tn) (lognot (ash 1 5)))))
-
+      (inst mov rax object)
+      (inst shr rax gencgc-card-shift)
+      (inst and :dword rax card-index-mask)
+      (inst mov :byte (ea gc-card-table-reg-tn rax) 0)
       STORE
-      (inst mov rdi-tn (ea 24 rsp-tn))      ; object
-      (inst mov temp-reg-tn (ea 32 rsp-tn)) ; word index
-      (inst mov rax-tn (ea 40 rsp-tn))      ; newval
+      (inst mov rdi object)
+      (inst mov rdx word-index)
+      (inst mov rax newval)
       ;; set 'written' flag in the code header
-      (inst or :byte :lock (ea (- 3 other-pointer-lowtag) rdi-tn) #x40)
+      (inst or :byte :lock (ea (- 3 other-pointer-lowtag) rdi) #x40)
       ;; store newval into object
-      (inst mov (ea (- other-pointer-lowtag) rdi-tn temp-reg-tn n-word-bytes)
-            rax-tn)))
-  (inst pop rdi-tn) ; restore
-  (inst pop rax-tn)
+      (inst mov (ea (- other-pointer-lowtag) rdi rdx n-word-bytes) rax))))
   (inst ret 24)) ; remove 3 stack args
 
-;;; Currently the only objects for which it is necessary to call TOUCH-GC-CARD
-;;; are those on varyobj pages. Therefore if no immobile-space feature, skip it.
-;;; This is not the situation for pages of code, where we manually toggle dynamic space
-;;; card marks so that when recording code coverage we don't incur the cost of a kernel
-;;; signal on a page that was otherwise untouched.
-#+sb-assembling
-(define-assembly-routine (touch-gc-card (:return-style :none)) ()
-  ;; stack: ret-pc, object
-  #+immobile-space
-  (progn
-   (ensure-thread-base-tn-loaded)
-   (inst mov temp-reg-tn (ea 8 rsp-tn))
-   (inst sub temp-reg-tn (thread-slot-ea thread-varyobj-space-addr-slot))
-   (inst shr temp-reg-tn (integer-length (1- immobile-card-bytes)))
-   (inst cmp temp-reg-tn (thread-slot-ea thread-varyobj-card-count-slot))
-   (inst jmp :ae DONE)
+;;; These are trampolines, but they benefit from not being in the 'tramps' file
+;;; because they'll automatically get a vop and an assembly routine this way,
+;;; where tramps only get the assembly routine.
+(define-assembly-routine (update-object-layout
+                          (:policy :fast-safe)
+                          (:translate update-object-layout)
+                          (:return-style :raw))
+    ((:arg x (descriptor-reg) rdx-offset)
+     (:res r (descriptor-reg) rdx-offset))
+  (progn x r)
+  (with-registers-preserved (lisp :except rdx-tn)
+    (call-static-fun 'update-object-layout 1)))
 
-   (inst push rax-tn)
-   (inst mov rax-tn (thread-slot-ea thread-varyobj-card-marks-slot))
-   (inst bts :dword :lock (ea rax-tn) temp-reg-tn)
-   (inst pop rax-tn))
-  DONE
-  (inst ret 8)) ; remove 1 stack arg
+(define-assembly-routine (sb-impl::install-hash-table-lock
+                          (:policy :fast-safe)
+                          (:translate sb-impl::install-hash-table-lock)
+                          (:return-style :raw))
+    ((:arg x (descriptor-reg) rdx-offset)
+     (:res r (descriptor-reg) rdx-offset))
+  (progn x r)
+  (with-registers-preserved (lisp :except rdx-tn)
+    (call-static-fun 'sb-impl::install-hash-table-lock 1)))
+
+;;; From a perspective of reducing code bloat, this asm routine does not merit
+;;; being one at all. A vop would suffice, because it's a single-use vop.
+;;; However the WITH-REGISTERS-PRESERVED macro seems to think that it is utilized
+;;; only by asm code in *ASSEMBLER-ROUTINES*. I had trouble with it otherwise.
+(define-assembly-routine (alloc-code (:return-style :raw))
+    ((:arg arg (signed-reg) rdx-offset)
+     (:res res (descriptor-reg) rdx-offset))
+  (with-registers-preserved (c :except rdx-tn)
+    (inst mov rdi-tn arg)
+    (pseudo-atomic ()
+      (inst call (make-fixup "alloc_code_object" :foreign)))
+    (move res rax-tn)))
