@@ -32,7 +32,7 @@
         (room-info-name info)))
 
 (defconstant tiny-boxed-size-mask #xFF)
-(defun !compute-room-infos ()
+(defun compute-room-infos ()
   (let ((infos (make-array 256 :initial-element nil))
         (default-size-mask (mask-field (byte 23 0) -1)))
     (dolist (obj *primitive-objects*)
@@ -47,11 +47,11 @@
                  (symbol (make-room-info tiny-boxed-size-mask name :other))
                  (t      (make-room-info default-size-mask name :other)))))))
 
-    (let ((info (make-room-info default-size-mask 'array-header :other)))
+    (let ((info (make-room-info tiny-boxed-size-mask 'array-header :other)))
       (dolist (code (list #+sb-unicode complex-character-string-widetag
                           complex-base-string-widetag simple-array-widetag
                           complex-bit-vector-widetag complex-vector-widetag
-                          complex-array-widetag complex-vector-nil-widetag))
+                          complex-array-widetag))
         (setf (svref infos code) info)))
 
     (setf (svref infos bignum-widetag)
@@ -71,6 +71,7 @@
         (when (saetp-specifier saetp) ;; SIMPLE-ARRAY-NIL is a special case.
           (setf (svref infos (saetp-typecode saetp)) saetp))))
 
+    ;; This one is here for completeness only- LENGTH does not imply size.
     (setf (svref infos simple-array-nil-widetag)
           (make-room-info 0 'simple-array-nil :other))
 
@@ -110,7 +111,8 @@
 
     infos))
 
-(define-load-time-global *room-info* (!compute-room-infos))
+(define-load-time-global *room-info* (compute-room-infos))
+(declaim (type (simple-vector 256) *room-info*))
 
 (defconstant-eqx +heap-spaces+
   '((:dynamic   "Dynamic space"   dynamic-usage)
@@ -186,71 +188,8 @@
                (ash header-word (- hash-slot-present-flag))
                1))))
 
-;;; * If symbols are 7 words incl header, as they are on 32-bit w/threads,
-;;;   then the part of NIL that manifests as symbol slots consumes 6 words
-;;;   (less the header) because NIL's symbol widetag precedes it by 1 word.
-;;;   So, there are 6 post-header words and 2 pre-header: one containing
-;;;   the widetag (not that it is ever read), and one 0 word.
-;;; * If symbols are 6 words incl header, as they are on 64-bit and
-;;;   32-bit w/o threads, then the symbol-like part of NIL is 5 words,
-;;;   which aligns up to 6, plus the two pre-header words.
-;;; So either way, NIL's alignment padding makes it come out to 8 words in total.
-(defconstant sizeof-nil-in-words (+ 2 (sb-int:align-up (1- sb-vm:symbol-size) 2)))
-
-;;; It's unclear to me whether reinventing sizetab[] in lisp is strictly
-;;; an improvement over doing the obvious:
-;;;  (WITH-PINNED-OBJECTS (object) (alien-funcall "ext_lispboj_size" ...))
-;;; It certainly doesn't feel that it's better, but it has significantly less overhead
-;;; at least on cheneygc if not also on the precise gencgc platforms.
-;;; But see WITH-PINNED-OBJECT-ITERATOR in 'target-hash-table' which offers
-;;; a way to mitigate the need to bind/unbind a special var when doing a massive
-;;; numbers of WITH-PINNED-OBJECT operations.
-;;; FIXME: export this from SB-EXT
-(defun primitive-object-size (object)
-  "Return number of bytes of heap or stack directly consumed by OBJECT"
-  (if (is-lisp-pointer (get-lisp-obj-address object))
-      (let ((words
-              (typecase object
-                (cons 2)
-                (instance (1+ (instance-length object)))
-                (function
-                 (when (= (fun-subtype object) simple-fun-widetag)
-                   (return-from primitive-object-size
-                     (primitive-object-size (fun-code-header object))))
-                 (1+ (get-closure-length object)))
-                (null sizeof-nil-in-words)
-                (code-component
-                 (return-from primitive-object-size (code-object-size object)))
-                (fdefn 4) ; no length stored in the header
-                ;; Anything else is an OTHER pointer.
-                (t
-                 (let ((room-info
-                         (aref *room-info* (%other-pointer-widetag object))))
-                   (if (arrayp object)
-                       (cond ((array-header-p object)
-                              (+ array-dimensions-offset (array-rank object)))
-                             ((simple-array-nil-p object) 2)
-                             (t
-                              (let* ((length (+ (length object)
-                                                (saetp-n-pad-elements room-info)))
-                                     (n-bits (saetp-n-bits room-info))
-                                     (alignment-pad (floor 7 n-bits))
-                                     ;; This math confuses me. So there's a self-test
-                                     ;; against the C code which is known good.
-                                     (n-data-octets (if (>= n-bits 8)
-                                                        (* length (ash n-bits -3))
-                                                        (ash (* (+ length alignment-pad)
-                                                                n-bits)
-                                                             -3))))
-                                (+ (ceiling n-data-octets n-word-bytes) ; N data words
-                                   vector-data-offset))))
-                       ;; Other things (symbol, value-cell, etc)
-                       ;; don't have a sizer, so use GET-HEADER-DATA
-                       (1+ (logand (get-header-data object)
-                                   (room-info-mask room-info)))))))))
-        (* (align-up words 2) n-word-bytes))
-      0))
-
+;;; Macros not needed after this file (and avoids a redefinition warning this way)
+(eval-when (:compile-toplevel)
 (defmacro widetag@baseptr (sap)
   #+big-endian `(sap-ref-8 ,sap ,(1- n-word-bytes))
   #+little-endian `(sap-ref-8 ,sap 0))
@@ -259,7 +198,7 @@
   `(%make-lisp-obj
     (logior (sap-int ,sap)
             (logand (deref (extern-alien "widetag_lowtag" (array char 256)) ,widetag)
-                    lowtag-mask))))
+                    lowtag-mask)))))
 
 ;;; This uses the funny fixnum representation of ADDRESS. I'd like to change this
 ;;; to take a SAP but god forbid people are using it?
@@ -307,25 +246,6 @@
      #-sb-devel
      (aver (sap= start end)))))
 
-;;; Test the sizing function ASAP, because if it's broken, then MAP-ALLOCATED-OBJECTS
-;;; is too, and then creating the initial core will crash because of the various heap
-;;; traversals performed in SAVE-LISP-AND-DIE. Don't delay a crash.
-(dovector (saetp *specialized-array-element-type-properties*)
-  (let ((length 0) (et (saetp-specifier saetp)))
-    (loop (let* ((array (make-array length :element-type et))
-                 (size-from-lisp (primitive-object-size array))
-                 (size-from-c
-                  (with-pinned-objects (array)
-                    (alien-funcall (extern-alien "ext_lispobj_size" (function unsigned unsigned))
-                                   (logandc2 (get-lisp-obj-address array) lowtag-mask)))))
-            (unless (= size-from-lisp size-from-c)
-              (bug "size calculation mismatch on ~S" array))
-            ;; Stop after enough trials to hit all the edge case
-            (when (or (>= size-from-lisp (* 8 sb-vm:n-word-bytes))
-                      (and (eq et nil) (>= length 4))) ; always 2 words
-              (return))
-            (incf length)))))
-
 ;;; Access to the GENCGC page table for better precision in
 ;;; MAP-ALLOCATED-OBJECTS
 #+gencgc
@@ -340,13 +260,12 @@
               (start #+64-bit (unsigned 32) #-64-bit signed)
               ;; On platforms with small enough GC pages, this field
               ;; will be a short. On platforms with larger ones, it'll
-              ;; be an int.
-              ;; Measured in bytes; the low bit has to be masked off.
-              (bytes-used (unsigned
-                           #.(if (typep gencgc-card-bytes '(unsigned-byte 16))
+              ;; be an int. It should probably never be an int.
+              (words-used (unsigned
+                           #.(if (typep gencgc-page-words '(unsigned-byte 16))
                                  16
                                  32)))
-              (flags (unsigned 8))
+              (flags (unsigned 8)) ; in C this is {type, need_zerofill, pinned}
               (gen (signed 8))))
   #+immobile-space
   (progn
@@ -442,10 +361,10 @@ We could try a few things to mitigate this:
              (multiple-value-bind (start end) (%space-bounds space)
                (declare (ignore start))
                (funcall fun nil symbol-widetag (* sizeof-nil-in-words n-word-bytes))
-               (let ((start (+ (logandc2 sb-vm:nil-value sb-vm:lowtag-mask)
-                               (ash (- sizeof-nil-in-words 2) sb-vm:word-shift))))
+               (let ((start (+ (logandc2 nil-value lowtag-mask)
+                               (ash (- sizeof-nil-in-words 2) word-shift))))
                  (map-objects-in-range fun
-                                       (ash start (- sb-vm:n-fixnum-tag-bits))
+                                       (ash start (- n-fixnum-tag-bits))
                                        end))))
 
             ((:read-only #-gencgc :dynamic)
@@ -492,7 +411,7 @@ We could try a few things to mitigate this:
        ;; in order to determine what regions contain objects.
 
        ;; We explicitly presume that any pages in an allocation region
-       ;; that are in-use have a BYTES-USED of GENCGC-CARD-BYTES
+       ;; that are in-use have a BYTES-USED of GENCGC-PAGE-BYTES
        ;; (indicating a full page) or an otherwise-valid BYTES-USED.
        ;; We also presume that the pages of an open allocation region
        ;; after the first page, and any pages that are unallocated,
@@ -500,7 +419,7 @@ We could try a few things to mitigate this:
 
        ;; Our procedure is to scan forward through the page table,
        ;; maintaining an "end pointer" until we reach a page where
-       ;; BYTES-USED is not GENCGC-CARD-BYTES or we reach
+       ;; BYTES-USED is not GENCGC-PAGE-BYTES or we reach
        ;; NEXT-FREE-PAGE.  We then MAP-OBJECTS-IN-RANGE if the range
        ;; is not empty, and proceed to the next page (unless we've hit
        ;; NEXT-FREE-PAGE).
@@ -510,7 +429,7 @@ We could try a few things to mitigate this:
        ;; prove to be an issue with concurrent systems, or with
        ;; spectacularly poor timing for closing an allocation region
        ;; in a single-threaded system.
-  (close-current-gc-region)
+  (close-thread-alloc-region)
   (do ((initial-next-free-page next-free-page)
        (base (int-sap (current-dynamic-space-start)))
        (start-page 0)
@@ -519,28 +438,27 @@ We could try a few things to mitigate this:
       ((> start-page initial-next-free-page))
          ;; The type constraint on page indices is probably too generous,
          ;; but it does its job of producing efficient code.
-    (declare (type (integer 0 (#.(/ (ash 1 n-machine-word-bits) gencgc-card-bytes)))
+    (declare (type (integer 0 (#.(/ (ash 1 n-machine-word-bits) gencgc-page-bytes)))
                    start-page end-page))
     (setq end-page start-page)
     (loop (setq end-page-bytes-used
-                ;; The low bit of bytes-used is the need-to-zero flag.
-                (logandc1 1 (slot (deref page-table end-page) 'bytes-used)))
+                (ash (slot (deref page-table end-page) 'words-used) word-shift))
           ;; See 'page_ends_contiguous_block_p' in gencgc.c
-          (when (or (< end-page-bytes-used gencgc-card-bytes)
+          (when (or (< end-page-bytes-used gencgc-page-bytes)
                     (= (slot (deref page-table (1+ end-page)) 'start) 0))
             (return))
           (incf end-page))
-    (let ((start (sap+ base (truly-the sb-vm:signed-word
-                                       (logand (* start-page gencgc-card-bytes)
+    (let ((start (sap+ base (truly-the signed-word
+                                       (logand (* start-page gencgc-page-bytes)
                                                most-positive-word))))
-          (end (sap+ base (truly-the sb-vm:signed-word
-                                     (logand (+ (* end-page gencgc-card-bytes)
+          (end (sap+ base (truly-the signed-word
+                                     (logand (+ (* end-page gencgc-page-bytes)
                                                 end-page-bytes-used)
                                              most-positive-word)))))
       (when (sap> end start)
-        ;; The bits in the 5-bit flag field have fixed positions,
+        ;; The bits in the 6-bit 'type' field have fixed positions,
         ;; but the position of the field itself depends on endianness.
-        (let ((flags (ldb #+little-endian (byte 5 0) #+big-endian (byte 5 3)
+        (let ((flags (ldb (byte 6 (+ #+big-endian 2))
                           (slot (deref page-table start-page) 'flags))))
           ;; The GEN slot is declared as (SIGNED 8) which does not satisfy the
           ;; type restriction on the first argument to LOGBITP.
@@ -555,27 +473,15 @@ We could try a few things to mitigate this:
                                   (< start-page initial-next-free-page))))))
     (setq start-page (1+ end-page))))
 
-;; Start with a Lisp rendition of ensure_region_closed() on the active
-;; thread's region, since users are often surprised to learn that a
-;; just-consed object can't necessarily be seen by MAP-ALLOCATED-OBJECTS.
+;; Users are often surprised to learn that a just-consed object can't
+;; necessarily be seen by MAP-ALLOCATED-OBJECTS, so close the region
+;; to update the page table.
 ;; Since we're in WITHOUT-GCING, there can be no interrupts.
-;; This is probably better than calling MAP-OBJECTS-IN-RANGE on the
-;; thread's region, because then we might visit some page twice
-;; by doing that.
-;; (And seeing small consing by other threads is hopeless either way)
-(defun close-current-gc-region ()
-  #+gencgc
-  (let ((region-struct
-          #+sb-thread (sap+ (sb-thread::current-thread-sap)
-                            (ash thread-alloc-region-slot word-shift))
-          ;; If no threads, the alloc_region struct is in static space.
-          #-sb-thread (int-sap boxed-region)))
-    ;; The 'start_addr' field is at word index 3 of the structure (see gencgc-alloc-region.h).
-    (unless (eql (sap-ref-word region-struct (ash 3 word-shift)) 0)
-      (alien-funcall (extern-alien "gc_close_region" (function void system-area-pointer int))
-                     region-struct
-                     1))
-    nil))
+;; Moreover it's probably not safe in the least to walk any thread's
+;; allocation region, unless the observer and observed aren't consing.
+(defun close-thread-alloc-region ()
+  #+gencgc (alien-funcall (extern-alien "close_current_thread_tlab" (function void)))
+  nil)
 
 ;;;; MEMORY-USAGE
 
@@ -782,9 +688,9 @@ We could try a few things to mitigate this:
                  (eql type funcallable-instance-widetag))
          (incf total-objects)
          (let* ((layout (if (eql type funcallable-instance-widetag)
-                            (%fun-layout obj)
-                            (%instance-layout obj)))
-                (classoid (layout-classoid layout))
+                            (%fun-wrapper obj)
+                            (%instance-wrapper obj)))
+                (classoid (wrapper-classoid layout))
                 (found (ensure-gethash classoid totals (cons 0 0)))
                 (size size))
            (declare (fixnum size))
@@ -922,19 +828,6 @@ We could try a few things to mitigate this:
 
 ;;;; LIST-ALLOCATED-OBJECTS, LIST-REFERENCING-OBJECTS
 
-(defvar *ignore-after* nil)
-
-(defun valid-obj (space x)
-  (or (not (eq space :dynamic))
-      ;; this test looks bogus if the allocator doesn't work linearly,
-      ;; which I suspect is the case for GENCGC.  -- CSR, 2004-06-29
-      (< (get-lisp-obj-address x) (get-lisp-obj-address *ignore-after*))))
-
-(defun maybe-cons (space x stuff)
-  (if (valid-obj space x)
-      (cons x stuff)
-      stuff))
-
 (defun list-allocated-objects (space &key type larger smaller count
                                      test)
   (declare (type (or (eql :all) spaces) space)
@@ -1054,12 +947,9 @@ We could try a few things to mitigate this:
          ;; These two are in fact generally the most frequently occurring type.
          ,.(make-case 'cons `(car ,obj) `(cdr ,obj))
          ,.(make-case* 'instance
-            `(let ((.l. (%instance-layout ,obj)))
-               ;; Though we've bound %INSTANCE-LAYOUT to a variable,
-               ;; pass the form %INSTANCE-LAYOUT to functoid
-               ;; in case it wants to examine the form
+            `(progn
                (,functoid (%instance-layout ,obj) ,@more)
-               (do-instance-tagged-slot (.i. ,obj :layout .l. :pad nil)
+               (do-instance-tagged-slot (.i. ,obj nil)
                  (,functoid (%instance-ref ,obj .i.) ,@more))))
          (function
           (typecase ,obj
@@ -1071,7 +961,7 @@ We could try a few things to mitigate this:
                   ;; if functoid is a macro that does nothing.
                   (,functoid .o. ,@more)))
             ,.(make-case* 'funcallable-instance
-               `(let ((.l. (%fun-layout ,obj)))
+               `(progn
                   ;; As for INSTANCE, allow the functoid to see the access form
                   (,functoid (%fun-layout ,obj) ,@more)
                   (,functoid (%funcallable-instance-fun ,obj) ,@more)
@@ -1081,7 +971,7 @@ We could try a few things to mitigate this:
                   ;; both tricky and unnecessary to generalize iteration.
                   ;; So just hardcode the few cases that exist.
                   #+compact-instance-header
-                  (ecase (layout-bitmap .l.)
+                  (ecase (wrapper-bitmap (%fun-wrapper ,obj))
                     (-1 ; external trampoline, all slots are tagged
                      ;; In this case, the trampoline word is scanned, with no ill effect.
                      (loop for .i. from 0
@@ -1095,7 +985,7 @@ We could try a few things to mitigate this:
                      (,functoid (%funcallable-instance-info ,obj 0) ,@more)))
                   #-compact-instance-header
                   (progn
-                    (aver (eql (layout-bitmap .l.) -4))
+                    (aver (eql (wrapper-bitmap (%fun-wrapper ,obj)) -4))
                     ;;            v ----trampoline
                     ;; = #b1...1100
                     ;;           ^----- layout
@@ -1122,7 +1012,7 @@ We could try a few things to mitigate this:
             ,.(make-case 'array)
             ,.(make-case* 'symbol
                `(,functoid (%primitive sb-c:fast-symbol-global-value ,obj) ,@more)
-               `(,functoid (symbol-info ,obj) ,@more)
+               `(,functoid (symbol-%info ,obj) ,@more)
                `(,functoid (symbol-name ,obj) ,@more)
                `(,functoid (symbol-package ,obj) ,@more)
                `(when (symbol-extra-slot-p ,obj)
@@ -1165,6 +1055,7 @@ We could try a few things to mitigate this:
 ;;; code-components are considered to reference their embedded
 ;;; simple-funs for this purpose; if THIS is a simple-fun, it is ignored.
 (defun references-p (this that)
+  (declare (optimize (sb-c::aref-trapping 0)))
   (macrolet ((test (x) `(when (eq ,x that) (go win))))
     (tagbody
        (do-referenced-object (this test)
@@ -1194,16 +1085,16 @@ We could try a few things to mitigate this:
 (defun map-referencing-objects (fun space object)
   (declare (type (or (eql :all) spaces) space))
   (declare (dynamic-extent fun))
-  (unless *ignore-after*
-    (setq *ignore-after* (cons 1 2)))
-  (let ((fun (%coerce-callable-to-fun fun)))
+  (let (list)
     (map-allocated-objects
      (lambda (referer widetag size)
        (declare (ignore widetag size))
-       (when (and (valid-obj space referer) ; semi-bogus!
+       ;; Don't count a self-reference as a reference
+       (when (and (neq referer object)
                   (references-p referer object))
-         (funcall fun referer)))
-     space)))
+         (push referer list)))
+     space)
+    (mapc (%coerce-callable-to-fun fun) list)))
 
 (defun list-referencing-objects (space object)
   (collect ((res))
@@ -1279,10 +1170,11 @@ We could try a few things to mitigate this:
                                 &aux (n-code-bytes 0)
                                      (total-pages next-free-page)
                                      (pages
-                                      (make-array total-pages :element-type 'bit)))
+                                      (make-array total-pages :element-type 'bit
+                                                  :initial-element 0)))
   (flet ((dump-page (page-num)
            (format stream "~&Page ~D~%" page-num)
-           (let ((where (+ dynamic-space-start (* page-num gencgc-card-bytes)))
+           (let ((where (+ dynamic-space-start (* page-num gencgc-page-bytes)))
                  (seen-filler nil))
              (loop
                (let* ((obj (let ((sap (int-sap where)))
@@ -1301,7 +1193,7 @@ We could try a few things to mitigate this:
                  (loop for index from page-num to (find-page-index (1- where))
                        do (setf (sbit pages index) 1)))
                (let ((next-page (find-page-index where)))
-                 (cond ((= (logand where (1- gencgc-card-bytes)) 0)
+                 (cond ((= (logand where (1- gencgc-page-bytes)) 0)
                         (format stream "~&-- END OF PAGE --~%")
                         (return next-page))
                        ((eq next-page page-num))
@@ -1314,7 +1206,7 @@ We could try a few things to mitigate this:
                      (setq i (dump-page i))
                      (incf i)))))
     (let* ((n-pages (count 1 pages))
-           (tot (* n-pages gencgc-card-bytes))
+           (tot (* n-pages gencgc-page-bytes))
            (waste (- tot n-code-bytes)))
       (format t "~&Used-bytes=~D Pages=~D Waste=~D (~F%)~%"
               n-code-bytes n-pages waste
@@ -1339,17 +1231,18 @@ We could try a few things to mitigate this:
 
 #+gencgc
 (defun generation-of (object)
-  (let* ((addr (get-lisp-obj-address object))
-         (page (find-page-index addr)))
-    (cond ((>= page 0) (slot (deref page-table page) 'gen))
-          #+immobile-space
-          ((immobile-space-addr-p addr)
-           ;; SIMPLE-FUNs don't contain a generation byte
-           (when (simple-fun-p object)
-             (setq addr (get-lisp-obj-address (fun-code-header object))))
-           (let ((sap (int-sap (logandc2 addr lowtag-mask))))
-             (logand (if (fdefn-p object) (sap-ref-8 sap 1) (sap-ref-8 sap 3))
-                     #xF))))))
+  (with-pinned-objects (object)
+    (let* ((addr (get-lisp-obj-address object))
+           (page (find-page-index addr)))
+      (cond ((>= page 0) (slot (deref page-table page) 'gen))
+            #+immobile-space
+            ((immobile-space-addr-p addr)
+             ;; SIMPLE-FUNs don't contain a generation byte
+             (when (simple-fun-p object)
+               (setq addr (get-lisp-obj-address (fun-code-header object))))
+             (let ((sap (int-sap (logandc2 addr lowtag-mask))))
+               (logand (if (fdefn-p object) (sap-ref-8 sap 1) (sap-ref-8 sap 3))
+                       #xF)))))))
 
 ;;; Show objects in a much simpler way than print-allocated-objects.
 ;;; Probably don't use this for generation 0 of dynamic space. Other spaces are ok.
@@ -1367,9 +1260,12 @@ We could try a few things to mitigate this:
 ;;; Unfortunately this is a near total copy of the test in gc.impure.lisp
 (defun !ensure-genesis-code/data-separation ()
   #+gencgc
-  (let* ((n-bits (+ next-free-page 10))
-         (code-bits (make-array n-bits :element-type 'bit))
-         (data-bits (make-array n-bits :element-type 'bit))
+  (let* ((n-bits
+          (progn
+            (close-thread-alloc-region)
+            (+ next-free-page 50)))
+         (code-bits (make-array n-bits :element-type 'bit :initial-element 0))
+         (data-bits (make-array n-bits :element-type 'bit :initial-element 0))
          (total-code-size 0))
     (map-allocated-objects
      (lambda (obj type size)
@@ -1385,6 +1281,20 @@ We could try a few things to mitigate this:
                                  data-bits)
                                 (t
                                  code-bits))))
+         ;; If Lisp allocates to cons pages, then this check is always valid.
+         ;; If not, then at first glance we could weaken the check to allow
+         ;; gen0 conses on MIXED pages, but even that is not enough- pinned conses
+         ;; will promote but keep their MIXED page type. So don't bother with this.
+         #+use-cons-region
+         (let* ((flags (slot (deref page-table (find-page-index obj-addr)) 'flags))
+                (type (ldb (byte 5 (+ #+big-endian 3)) flags))
+                (ok (if (consp obj)
+                        (or (= type #b101) ; PAGE_TYPE_CONS
+                            (and (eq (car obj) 0) (eq (cdr obj) 0)))
+                        (/= type #b101))))
+           (unless ok
+             (error "Object @ ~x (gen~D) is on page-type ~b~%"
+                    obj-addr (generation-of obj) type)))
          ;; This is not the most efficient way to update the bit arrays,
          ;; but the simplest and clearest for sure. (The loop could avoided
          ;; if the current page is the same as the previously seen page)
@@ -1395,7 +1305,7 @@ We could try a few things to mitigate this:
                do (cond ((= (sbit other-array index) 1)
                          (format t "~&broken on page index ~d base ~x~%"
                                  index
-                                 (+ dynamic-space-start (* index gencgc-card-bytes)))
+                                 (+ dynamic-space-start (* index gencgc-page-bytes)))
                          (alien-funcall (extern-alien "ldb_monitor" (function void))))
                         (t
                          (setf (sbit array index) 1))))))
@@ -1426,31 +1336,38 @@ We could try a few things to mitigate this:
     (with-pinned-objects (obj)
       (dotimes (i count)
         (let ((word (sap-ref-word (int-sap addr) (ash i word-shift))))
-          (multiple-value-bind (lispobj ok)
+          (multiple-value-bind (lispobj ok fmt)
               (cond ((and (typep thing 'code-component)
                           (< 1 i (code-header-words thing)))
                      (values (code-header-ref thing i) t))
+                    #+compact-symbol
+                    ((and (typep thing '(and symbol (not null)))
+                          (= i symbol-name-slot))
+                     (values (list (sb-impl::symbol-package-id thing)
+                                   (symbol-name thing))
+                             t
+                             "{~{~A,~S~}}"))
                     (decode
                      (make-lisp-obj word nil)))
             (let ((*print-lines* 1)
                   (*print-pretty* t))
-              (format t "~x: ~v,'0x~:[~; = ~S~]~%"
+              (format t "~x: ~v,'0x~:[~; = ~@?~]~%"
                       (+ addr (ash i word-shift))
                       (* 2 n-word-bytes)
-                      word ok lispobj))))))))
+                      word ok (or fmt "~S") lispobj))))))))
 #+sb-thread
 (defun show-tls-map ()
   (let ((list
-         (sort (sb-vm::list-allocated-objects
+         (sort (list-allocated-objects
                 :all
-                :type sb-vm:symbol-widetag
+                :type symbol-widetag
                 :test (lambda (x) (plusp (sb-kernel:symbol-tls-index x))))
                #'<
                :key #'sb-kernel:symbol-tls-index))
         (prev 0))
     (dolist (x list)
-      (let ((n  (ash (sb-kernel:symbol-tls-index x) (- sb-vm:word-shift))))
-        (when (and (> n sb-vm::primitive-thread-object-length)
+      (let ((n  (ash (sb-kernel:symbol-tls-index x) (- word-shift))))
+        (when (and (> n primitive-thread-object-length)
                    (> n (1+ prev)))
           (format t "(unused)~%"))
         (format t "~5d = ~s~%" n x)
@@ -1462,43 +1379,53 @@ We could try a few things to mitigate this:
          (let ((*print-level* 2) (*print-lines* 1))
            (format t "~x ~s~%" (get-lisp-obj-address obj) obj))))
 (defun print-all-code ()
-  (walk-dynamic-space #'print-it #x7f #b11 #b11))
+  (walk-dynamic-space #'print-it #x7f #b01111 #b00111))
 (defun print-large-code ()
-  (walk-dynamic-space #'print-it #x7f #b10011 #b10011))
+  (walk-dynamic-space #'print-it #x7f #b11111 #b10111))
 (defun print-large-unboxed ()
-  (walk-dynamic-space #'print-it #x7f #b10011 #b10010))
+  (walk-dynamic-space #'print-it #x7f #b11111 #b10010))
 ;;; Use this only if you know what you're doing. It can fail because a page
 ;;; that needs to continue onto the next page will cause the "overrun" check
 ;;; to fail.
 (defun print-page-contents (page)
   (let* ((start
-          (+ (current-dynamic-space-start) (* sb-vm:gencgc-card-bytes page)))
+          (+ (current-dynamic-space-start) (* gencgc-page-bytes page)))
          (end
-          (+ start sb-vm:gencgc-card-bytes)))
+          (+ start gencgc-page-bytes)))
     (map-objects-in-range #'print-it (%make-lisp-obj start) (%make-lisp-obj end)))))
 
-(defun code-from-serialno (serial)
-  (dx-flet ((visit (obj type size)
+(defun map-code-objects (fun)
+  (dx-flet ((filter (obj type size)
               (declare (ignore size))
-              #+gencgc (aver (= type sb-vm:code-header-widetag))
-              (when (= (%code-serialno obj) serial)
-                (return-from code-from-serialno obj))))
-    #+cheneygc (map-allocated-objects #'visit :all)
+              (when (= type code-header-widetag)
+                (funcall fun obj)))
+            (nofilter (obj type size)
+              (declare (ignore type size))
+              (funcall fun obj)))
+    #+cheneygc (map-allocated-objects #'filter :all)
     #+gencgc
-    (progn
+    (without-gcing
       #+immobile-code
-      (map-objects-in-range #'visit
+      (map-objects-in-range #'nofilter
                             (ash varyobj-space-start (- n-fixnum-tag-bits))
                             (%make-lisp-obj (sap-int *varyobj-space-free-pointer*)))
-      (walk-dynamic-space #'visit
+      (alien-funcall (extern-alien "close_code_region" (function void)))
+      (walk-dynamic-space #'nofilter
                           #b1111111 ; all generations
-                          3 3))))
+                          #b111 #b111)))) ; type mask and constraint
+
+(export 'code-from-serialno)
+(defun code-from-serialno (serial)
+  (dx-flet ((visit (obj)
+              (when (= (%code-serialno obj) serial)
+                (return-from code-from-serialno obj))))
+    (map-code-objects #'visit)))
 
 (defun show-all-layouts ()
-  (let ((l (sb-vm::list-allocated-objects :all :test #'sb-kernel::layout-p))
+  (let ((l (list-allocated-objects :all :test #'sb-kernel::wrapper-p))
         zero trailing-raw trailing-tagged vanilla)
     (dolist (x l)
-      (let ((m (sb-kernel::layout-bitmap x)))
+      (let ((m (wrapper-bitmap x)))
         (cond ((eql m +layout-all-tagged+) (push x vanilla))
               ((eql m 0) (push x zero))
               ((minusp m) (push x trailing-tagged))
@@ -1506,42 +1433,61 @@ We could try a few things to mitigate this:
     (flet ((legend (newline str list)
              (when newline (terpri))
              (let ((s (format nil str (length list))))
-               (format t "~A~%~A~%" s (make-string (length s) :initial-element #\-))))
-           (name (x) (classoid-name (layout-classoid x))))
+               (format t "~A~%~A~%" s (make-string (length s) :initial-element #\-)))))
       (when zero
         (legend nil "Zero bitmap (~d):" zero)
-        (dolist (x zero) (format t "~a~%" (name x))))
+        (dolist (x zero) (format t "~a~%" (wrapper-classoid-name x))))
       (when trailing-raw
         (legend t "Trailing raw (~d):" trailing-raw)
         (dolist (x trailing-raw)
-          (let ((m (sb-kernel::layout-bitmap x)))
+          (let ((m (wrapper-bitmap x)))
             (format t "~30a 0...~v,'0b~%"
-                    (name x)
-                    (acond ((layout-info x) (1+ (dd-length it))) (t 32))
+                    (wrapper-classoid-name x)
+                    (acond ((wrapper-info x) (1+ (dd-length it))) (t 32))
                     m))))
       (when trailing-tagged
         (legend t "Trailing tagged (~d):" trailing-tagged)
         (dolist (x trailing-tagged)
-          (let ((m (sb-kernel::layout-bitmap x)))
+          (let ((m (wrapper-bitmap x)))
             (format t "~30a 1...~b~%"
-                    (name x)
-                    (acond ((layout-info x) (ldb (byte (dd-length it) 0) m))
+                    (wrapper-classoid-name x)
+                    (acond ((wrapper-info x) (ldb (byte (dd-length it) 0) m))
                            (t (ldb (byte 32 0) m)))))))
       (legend t "Default: (~d) [not shown]" vanilla))))
 
+#+ubsan
+(defun find-poisoned-vectors (&aux result)
+  (dolist (v (list-allocated-objects :all :type simple-vector-widetag)
+             result)
+    (when (dotimes (i (length v))
+            (declare (optimize (sb-c::aref-trapping 0)))
+            (let ((val (svref v i)))
+              (when (= (get-lisp-obj-address val) no-tls-value-marker-widetag)
+                (return t))))
+      (push (make-weak-pointer v) result)
+      (let* ((origin (vector-extra-data v))
+             (code (sb-di::code-header-from-pc (ash origin -3)))
+             (*print-array* nil))
+        (format t "g~d ~a ~a~%" (sb-kernel:generation-of v) v code)))))
+
 (in-package "SB-C")
 ;;; As soon as practical in warm build it makes sense to add
-;;; cold-allocation-point-fixups into the weak hash-table.
+;;; cold-allocation-patch-points into the weak hash-table.
 ;;; FIXME: I suspect that this wants to be just a weak vector
 ;;; (all code objects that have any allocation profiling compiled in),
 ;;; and not a hash-table, and that the list of fixups in the component
 ;;; can be attached to the debug info (in the manner of debug funs).
 ;;; When this was first implemented, weak-vectors weren't a thing. Maybe?
-(defvar *!cold-allocation-point-fixups*)
-(let ((hash-table (make-hash-table :test 'eq)))
-  ;; Group by code component
-  (loop for (code . offset) across *!cold-allocation-point-fixups*
-        do (push offset (gethash code hash-table)))
-  (dohash ((code list) hash-table)
-    (setf (gethash code *allocation-point-fixups*)
-          (convert-alloc-point-fixups code list))))
+(defvar *!cold-allocation-patch-point*)
+(loop for (code . points) in *!cold-allocation-patch-point*
+      do (setf (gethash code *allocation-patch-points*) points))
+
+(defun gctablesize (heap-size-gb page-size cards-per-page)
+  (let* ((card-size (/ page-size cards-per-page))
+         (heap-size (* heap-size-gb (expt 1024 3)))
+         (npages (/ heap-size page-size))
+         (ncards (/ heap-size card-size))
+         (pte-nbytes (* 8 npages)))
+    (format t " PTE bytes: ~8D~%" pte-nbytes)
+    (format t "card bytes: ~8D~%" ncards)
+    (format t "     total: ~8D~%" (+ pte-nbytes ncards))))
