@@ -40,36 +40,6 @@
 
 ;;;; miscellaneous types used both in the cross-compiler and on the target
 
-;;;; FIXME: The INDEX and LAYOUT-DEPTHOID definitions probably belong
-;;;; somewhere else, not "early-c", since they're after all not part
-;;;; of the compiler.
-
-;;; the type of LAYOUT-DEPTHOID and LAYOUT-LENGTH values.
-;;; Each occupies two bytes of the %BITS slot when possible,
-;;; otherwise a slot unto itself.
-(def!type layout-depthoid () '(integer -1 #x7FFF))
-(def!type layout-length () '(integer 0 #xFFFF))
-(def!type layout-bitmap ()
-  ;; FIXME: Probably should exclude negative bignum
-  #+compact-instance-header 'integer
-  #-compact-instance-header '(and integer (not (eql 0))))
-
-;;; An INLINEP value describes how a function is called. The values
-;;; have these meanings:
-;;;     NIL     No declaration seen: do whatever you feel like, but don't
-;;;             dump an inline expansion.
-;;; NOTINLINE  NOTINLINE declaration seen: always do full function call.
-;;;    INLINE  INLINE declaration seen: save expansion, expanding to it
-;;;             if policy favors.
-;;; MAYBE-INLINE
-;;;             Retain expansion, but only use it opportunistically.
-;;;             MAYBE-INLINE is quite different from INLINE. As explained
-;;;             by APD on #lisp 2005-11-26: "MAYBE-INLINE lambda is
-;;;             instantiated once per component, INLINE - for all
-;;;             references (even under #'without FUNCALL)."
-(deftype inlinep ()
-  '(member inline maybe-inline notinline nil))
-
 (defstruct (dxable-args (:constructor make-dxable-args (list))
                         (:predicate nil)
                         (:copier nil))
@@ -89,15 +59,17 @@
   (free-funs (make-hash-table :test 'equal) :read-only t :type hash-table)
   ;; These hashtables translate from constants to the LEAFs that
   ;; represent them.
-  ;; Table 1: one entry for each distinct constant (according to object identity)
-  (eq-constants (make-hash-table :test 'eq) :read-only t :type hash-table)
-  ;; Table 2: one hash-table entry per EQUAL constant,
+  ;; Table 1: one entry per named constant
+  (named-constants (make-hash-table :test 'eq) :read-only t :type hash-table)
+  ;; Table 2: one entry for each unnamed constant as compared by EQL
+  (eql-constants (make-hash-table :test 'eql) :read-only t :type hash-table)
+  ;; Table 3: one key per EQUAL constant,
   ;; with the caveat that lookups must discriminate amongst constants that
   ;; are EQUAL but not similar.  The value in the hash-table is a list of candidates
   ;; (#<constant1> #<constant2> ... #<constantN>) such that CONSTANT-VALUE
   ;; of each is EQUAL to the key for the hash-table entry, but dissimilar
   ;; from each other. Notably, strings of different element types can't be similar.
-  (similar-constants (make-hash-table :test 'equal) :read-only t :type hash-table))
+  (similar-constants (sb-fasl::make-similarity-table) :read-only t :type hash-table))
 (declaim (freeze-type ir1-namespace))
 
 (sb-impl::define-thread-local *ir1-namespace*)
@@ -120,12 +92,11 @@
 ;;; Bind this to a stream to capture various internal debugging output.
 (defvar *compiler-trace-output* nil)
 ;;; These are the default, but the list can also include
-;;; :pre-ir2-optimize and :symbolic-asm.
+;;; :pre-ir2-optimize, :symbolic-asm, and :sb-graph.
 (defvar *compile-trace-targets* '(:ir1 :ir2 :vop :symbolic-asm :disassemble))
 (defvar *constraint-universe*)
 (defvar *current-path*)
 (defvar *current-component*)
-(defvar *delayed-ir1-transforms*)
 #+sb-dyncount
 (defvar *dynamic-counts-tn*)
 (defvar *elsewhere-label*)
@@ -140,6 +111,9 @@
 (defvar *compile-object* nil)
 (defvar *location-context* nil)
 
+(defvar *handled-conditions* nil)
+(defvar *disabled-package-locks* nil)
+
 (defvar *stack-allocate-dynamic-extent* t
   "If true (the default), the compiler respects DYNAMIC-EXTENT declarations
 and stack allocates otherwise inaccessible parts of the object whenever
@@ -153,94 +127,17 @@ the stack without triggering overflow protection.")
 (declaim (type (member nil t :specified)
                *block-compile-default* *block-compile-argument*))
 
-;;; This lock is seized in the compiler, and related areas -- like the
-;;; classoid/layout/class system.
-;;; Assigning a literal object enables genesis to dump and load it
-;;; without need of a cold-init function.
-#-sb-xc-host
-(!define-load-time-global **world-lock** (sb-thread:make-mutex :name "World Lock"))
+;; Names seen which are defined to be hairy (i.e. non-EQ comparable)
+;; constants.
+(defvar *hairy-defconstants* '())
+(declaim (type list *hairy-defconstants*))
 
-#-sb-xc-host
-(define-load-time-global *static-linker-lock*
-    (sb-thread:make-mutex :name "static linker"))
-
-(defmacro with-world-lock (() &body body)
-  #+sb-xc-host `(progn ,@body)
-  #-sb-xc-host `(sb-thread:with-recursive-lock (**world-lock**) ,@body))
 
 ;;;; miscellaneous utilities
 
-;;; This is for "observers" who want to know if type names have been added.
-;;; Rather than registering listeners, they can detect changes by comparing
-;;; their stored nonce to the current nonce. Additionally the observers
-;;; can detect whether function definitions have occurred.
-#-sb-xc-host
-(progn (declaim (fixnum *type-cache-nonce*))
-       (!define-load-time-global *type-cache-nonce* 0))
-
-(defstruct (undefined-warning
-            (:print-object (lambda (x s)
-                             (print-unreadable-object (x s :type t)
-                               (prin1 (undefined-warning-name x) s))))
-            (:copier nil))
-  ;; the name of the unknown thing
-  (name nil :type (or symbol list))
-  ;; the kind of reference to NAME
-  (kind (missing-arg) :type (member :function :type :variable))
-  ;; the number of times this thing was used
-  (count 0 :type unsigned-byte)
-  ;; a list of COMPILER-ERROR-CONTEXT structures describing places
-  ;; where this thing was used. Note that we only record the first
-  ;; *UNDEFINED-WARNING-LIMIT* calls.
-  (warnings () :type list))
-(declaim (freeze-type undefined-warning))
-
-;;; Delete any undefined warnings for NAME and KIND. This is for the
-;;; benefit of the compiler, but it's sometimes called from stuff like
-;;; type-defining code which isn't logically part of the compiler.
-(declaim (ftype (function ((or symbol cons) keyword) (values))
-                note-name-defined))
-(defun note-name-defined (name kind)
-  #-sb-xc-host (atomic-incf *type-cache-nonce*)
-  ;; We do this BOUNDP check because this function can be called when
-  ;; not in a compilation unit (as when loading top level forms).
-  (when (boundp '*undefined-warnings*)
-    (let ((name (uncross name)))
-      (setq *undefined-warnings*
-            (delete-if (lambda (x)
-                         (and (equal (undefined-warning-name x) name)
-                              (eq (undefined-warning-kind x) kind)))
-                       *undefined-warnings*))))
-  (values))
-
-;;; to be called when a variable is lexically bound
-(declaim (ftype (function (symbol) (values)) note-lexical-binding))
-(defun note-lexical-binding (symbol)
-    ;; This check is intended to protect us from getting silently
-    ;; burned when we define
-    ;;   foo.lisp:
-    ;;     (DEFVAR *FOO* -3)
-    ;;     (DEFUN FOO (X) (+ X *FOO*))
-    ;;   bar.lisp:
-    ;;     (DEFUN BAR (X)
-    ;;       (LET ((*FOO* X))
-    ;;         (FOO 14)))
-    ;; and then we happen to compile bar.lisp before foo.lisp.
-  (when (looks-like-name-of-special-var-p symbol)
-    ;; FIXME: should be COMPILER-STYLE-WARNING?
-    (style-warn 'asterisks-around-lexical-variable-name
-                :format-control
-                "using the lexical binding of the symbol ~
-                 ~/sb-ext:print-symbol-with-prefix/, not the~@
-                 dynamic binding"
-                :format-arguments (list symbol)))
-  (values))
-
-;;; This is DEF!STRUCT so that when SB-C:DUMPABLE-LEAFLIKE-P invokes
-;;; SB-XC:TYPEP in make-host-2, it does not need need to signal PARSE-UNKNOWN
-;;; for each and every constant seen up until this structure gets defined.
-(def!struct (debug-name-marker (:print-function print-debug-name-marker)
-                               (:copier nil)))
+(defstruct (debug-name-marker (:print-function print-debug-name-marker)
+                              (:copier nil)))
+(declaim (freeze-type debug-name-marker))
 
 (defvar *debug-name-level* 4)
 (defvar *debug-name-length* 12)
@@ -284,6 +181,8 @@ the stack without triggering overflow protection.")
                  ((or symbol number string)
                   x)
                  (t
+                  ;; wtf?? This looks like a source of sensitivity to the cross-compiler host
+                  ;; in addition to which it seems generally a stupid idea.
                   (type-of x)))))
       (let ((name (list* type (walk thing) (when context (name-context)))))
         (when (legal-fun-name-p name)
@@ -395,7 +294,7 @@ the stack without triggering overflow protection.")
   (objmap-id-to-ir2block nil :type (or null id-array)) ; number -> IR2-BLOCK
   (objmap-id-to-tn       nil :type (or null id-array)) ; number -> TN
   (objmap-id-to-label    nil :type (or null id-array)) ; number -> LABEL
-  )
+  deleted-source-paths)
 (declaim (freeze-type compilation))
 
 (sb-impl::define-thread-local *compilation*)
