@@ -343,6 +343,9 @@
 (defknown %multiply-high (word word) word
     (movable foldable flushable))
 
+(defknown multiply-fixnums (fixnum fixnum) integer
+  (movable foldable flushable))
+
 (defknown %signed-multiply-high (sb-vm:signed-word sb-vm:signed-word) sb-vm:signed-word
     (movable foldable flushable))
 
@@ -1076,7 +1079,7 @@
 (defknown hash-table-rehash-threshold (hash-table) (single-float ($0.0) $1.0)
   (foldable flushable))
 (defknown hash-table-size (hash-table) index (flushable))
-(defknown hash-table-test (hash-table) symbol (foldable flushable))
+(defknown hash-table-test (hash-table) function-designator (foldable flushable))
 (defknown (sxhash psxhash) (t) hash-code (foldable flushable))
 (defknown hash-table-equalp (hash-table hash-table) boolean (foldable flushable))
 (defknown sb-impl::install-hash-table-lock (hash-table) sb-thread:mutex ())
@@ -1110,7 +1113,7 @@
                        (:displaced-to (or array null))
                        (:displaced-index-offset index))
     array (flushable))
-
+(defknown initial-contents-error (t t) nil)
 (defknown fill-data-vector (vector list sequence) vector ()
   :result-arg 0)
 
@@ -1271,7 +1274,21 @@
                      (inhibit-flushing index 0) (inhibit-flushing sequence-end nil)
                      (inhibit-flushing index 0) (inhibit-flushing sequence-end nil))
   boolean
-  (foldable flushable))
+    (foldable flushable))
+
+(defknown simple-base-string=
+    (simple-base-string simple-base-string
+                        (inhibit-flushing index 0) (inhibit-flushing sequence-end nil)
+                        (inhibit-flushing index 0) (inhibit-flushing sequence-end nil))
+    boolean
+    (foldable flushable))
+
+(defknown simple-character-string=
+    (simple-character-string simple-character-string
+                             (inhibit-flushing index 0) (inhibit-flushing sequence-end nil)
+                             (inhibit-flushing index 0) (inhibit-flushing sequence-end nil))
+    boolean
+    (foldable flushable))
 
 ;;;; from the "Eval" chapter:
 
@@ -1677,21 +1694,29 @@
    &key
 
    ;; ANSI options
-   (:output-file (or pathname-designator
-                     null
-                     ;; FIXME: This last case is a non-ANSI hack.
-                     (member t)))
+   (:output-file pathname-designator)
    (:verbose t)
    (:print t)
    (:external-format external-format-designator)
-   (:progress t)
 
    ;; extensions
+   (:progress t)
    (:trace-file t)
    (:block-compile t)
    (:entry-points list)
    (:emit-cfasl t))
   (values (or pathname null) boolean boolean))
+(defknown sb-c::compile-files
+    (cons &key (:output-file pathname-designator)
+               (:verbose t)
+               (:print t)
+               (:external-format external-format-designator)
+               (:progress t)
+               (:trace-file t)
+               (:block-compile t)
+               (:entry-points list)
+               (:emit-cfasl t))
+  (values (or pathname null) boolean boolean))
 
 (defknown (compile-file-pathname)
   (pathname-designator &key (:output-file (or pathname-designator
@@ -1807,12 +1832,13 @@
   (movable flushable always-translatable)
   :derive-type #'result-type-first-arg)
 
-(defknown %cleanup-point (&rest t) t)
+(defknown %cleanup-point (&rest t) t (reoptimize-when-unlinking))
 (defknown %special-bind (t t) t)
 (defknown %special-unbind (&rest symbol) t)
 (defknown %listify-rest-args (t index) list (flushable))
 (defknown %more-arg-context (t t) (values t index) (flushable))
-(defknown %more-arg (t index) t)
+(defknown %more-arg (t index) t (flushable))
+(defknown %more-keyword-pair (t fixnum) (values t t) (flushable))
 #+stack-grows-downward-not-upward
 ;;; FIXME: The second argument here should really be NEGATIVE-INDEX, but doing that
 ;;; breaks the build, and I cannot seem to figure out why. --NS 2006-06-29
@@ -1853,9 +1879,9 @@
 (defknown (%check-bound check-bound) (array index t) index
   (dx-safe))
 (defknown data-vector-ref (simple-array index) t
-  (foldable unsafely-flushable always-translatable))
+  (foldable flushable always-translatable))
 (defknown data-vector-ref-with-offset (simple-array fixnum fixnum) t
-  (foldable unsafely-flushable always-translatable))
+  (foldable flushable always-translatable))
 (defknown data-nil-vector-ref (simple-array index) nil
   (always-translatable))
 ;;; The lowest-level vector SET operators should not return a value.
@@ -1978,8 +2004,11 @@
 (defknown %scharset ((modifying simple-string) index character) character ())
 (defknown %set-symbol-value (symbol t) t ())
 (defknown (setf symbol-function) (function symbol) function ())
-(defknown %set-symbol-plist (symbol list) list ()
-  :derive-type #'result-type-last-arg)
+;; Does this really need a type deriver? It's inline, and returns its 1st arg,
+;; i.e. we know exactly what object it returns, which is more precise than
+;; just knowing the type.
+(defknown (setf symbol-plist) (list symbol) list ()
+  :derive-type #'result-type-first-arg)
 (defknown %setnth (unsigned-byte (modifying list) t) t ()
   :derive-type #'result-type-last-arg)
 (defknown %set-fill-pointer ((modifying complex-vector) index) index ()
@@ -1987,7 +2016,7 @@
 
 ;;;; ALIEN and call-out-to-C stuff
 
-(defknown %alien-funcall ((or string system-area-pointer) alien-type &rest *) *)
+(defknown %alien-funcall ((or string system-area-pointer) alien-type &rest t) *)
 
 ;; Used by WITH-PINNED-OBJECTS
 #+(or x86 x86-64)
@@ -2069,6 +2098,10 @@
 (defknown sb-kernel::gc-safepoint () (values) ())
 
 ;;;; atomic ops
+;;; the CAS functions are transformed to something else rather than "translated".
+;;; either way, they should not be called.
+(defknown (cas svref) (t t simple-vector index) t (always-translatable))
+(defknown (cas symbol-value) (t t symbol) t (always-translatable))
 (defknown %compare-and-swap-svref (simple-vector index t t) t
     ())
 (defknown (%compare-and-swap-symbol-value
