@@ -9,29 +9,21 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!VM")
+(in-package "SB-VM")
 
 ;;;; LIST and LIST*
-(define-vop (list-or-list*)
-  (:args (things :more t))
+(define-vop (list)
+  (:args (things :more t :scs (any-reg descriptor-reg zero null control-stack)))
   (:temporary (:scs (descriptor-reg)) ptr)
   (:temporary (:scs (descriptor-reg)) temp)
   (:temporary (:scs (descriptor-reg) :to (:result 0) :target result)
               res)
   (:temporary (:scs (non-descriptor-reg)) alloc-temp)
-  (:info num)
+  (:info star cons-cells)
   (:results (result :scs (descriptor-reg)))
-  (:variant-vars star)
-  (:policy :safe)
   (:node-var node)
   (:generator 0
-    (cond ((zerop num)
-           (move result null-tn))
-          ((and star (= num 1))
-           (move result (tn-ref-tn things)))
-          (t
-           (macrolet
-               ((maybe-load (tn)
+    (macrolet ((maybe-load (tn)
                   (once-only ((tn tn))
                     `(sc-case ,tn
                        ((any-reg descriptor-reg zero null)
@@ -39,11 +31,10 @@
                        (control-stack
                         (load-stack-tn temp ,tn)
                         temp)))))
-             (let* ((dx-p (node-stack-allocate-p node))
-                    (cons-cells (if star (1- num) num))
-                    (alloc (* (pad-data-block cons-size) cons-cells)))
-               (pseudo-atomic ()
-                 (allocation res alloc list-pointer-lowtag
+      (let ((dx-p (node-stack-allocate-p node))
+            (alloc (* (pad-data-block cons-size) cons-cells)))
+        (pseudo-atomic ()
+                 (allocation 'list alloc list-pointer-lowtag res
                              :stack-p dx-p
                              :temp-tn alloc-temp)
                  (move ptr res)
@@ -61,47 +52,10 @@
                              (maybe-load (tn-ref-tn (tn-ref-across things)))
                              null-tn)
                          ptr cons-cdr-slot list-pointer-lowtag))
-               (move result res)))))))
-
-(define-vop (list list-or-list*)
-  (:variant nil))
-
-(define-vop (list* list-or-list*)
-  (:variant t))
-
+        (move result res)))))
 
 ;;;; Special purpose inline allocators.
 
-(define-vop (allocate-code-object)
-  ;; BOXED is a count of words as a fixnum; it is therefore also a byte count
-  ;; as a raw value because n-fixnum-tag-bits = word-shift.
-  (:args (boxed :scs (any-reg))
-         (unboxed-arg :scs (any-reg)))
-  (:results (result :scs (descriptor-reg)))
-  (:temporary (:scs (non-descriptor-reg)) ndescr)
-  (:temporary (:scs (non-descriptor-reg)) size)
-  (:temporary (:scs (non-descriptor-reg)) unboxed)
-  (:generator 100
-    (inst srl unboxed unboxed-arg word-shift)
-    (inst add unboxed lowtag-mask)
-    (inst and unboxed (lognot lowtag-mask))
-    (pseudo-atomic ()
-      ;;
-      ;; This looks like another dreadful type pun. CSR - 2002-02-06
-      ;;
-      ;; Not any more, or not in that sense at least, because the
-      ;; "p/a bit is also the highest lowtag bit" assumption is now hidden
-      ;; in the allocation macro.  DFL - 2012-10-01
-      ;;
-      ;; Figure out how much space we really need and allocate it.
-      (inst add size boxed unboxed)
-      (allocation result size other-pointer-lowtag :temp-tn ndescr)
-      (inst sll ndescr boxed (- n-widetag-bits word-shift))
-      (inst or ndescr code-header-widetag)
-      (storew ndescr result 0 other-pointer-lowtag)
-      (storew unboxed-arg result code-code-size-slot other-pointer-lowtag)
-      (storew null-tn result code-debug-info-slot other-pointer-lowtag))))
-
 (define-vop (make-fdefn)
   (:args (name :scs (descriptor-reg) :to :eval))
   (:temporary (:scs (non-descriptor-reg)) temp)
@@ -126,7 +80,7 @@
     (let* ((size (+ length closure-info-offset))
            (alloc-size (pad-data-block size)))
       (pseudo-atomic ()
-        (allocation result alloc-size fun-pointer-lowtag
+        (allocation nil alloc-size fun-pointer-lowtag result
                     :stack-p stack-allocate-p
                     :temp-tn temp)
         (inst li temp (logior (ash (1- size) n-widetag-bits) closure-widetag))
@@ -162,30 +116,35 @@
 (define-vop (fixed-alloc)
   (:args)
   (:info name words type lowtag stack-allocate-p)
-  (:ignore name stack-allocate-p)
+  (:ignore name)
   (:results (result :scs (descriptor-reg)))
   (:temporary (:scs (non-descriptor-reg)) temp)
   (:generator 4
     (pseudo-atomic ()
-      (allocation result (pad-data-block words) lowtag :temp-tn temp)
-      (when type
-        (inst li temp (logior (ash (1- words) n-widetag-bits) type))
-        (storew temp result 0 lowtag)))))
+      (allocation nil (pad-data-block words) lowtag result :temp-tn temp
+                  :stack-p stack-allocate-p)
+        (inst li temp (compute-object-header words type))
+        (storew temp result 0 lowtag))))
 
 (define-vop (var-alloc)
   (:args (extra :scs (any-reg)))
   (:arg-types positive-fixnum)
-  (:info name words type lowtag)
-  (:ignore name)
+  (:info name words type lowtag stack-allocate-p)
+  (:ignore name stack-allocate-p)
   (:results (result :scs (descriptor-reg)))
   (:temporary (:scs (any-reg)) bytes)
   (:temporary (:scs (non-descriptor-reg)) header)
   (:temporary (:scs (non-descriptor-reg)) temp)
   (:generator 6
     (inst add bytes extra (* (1+ words) n-word-bytes))
-    (inst sll header bytes (- n-widetag-bits 2))
-    (inst add header header (+ (ash -2 n-widetag-bits) type))
-    (inst and bytes (lognot lowtag-mask))
+    (inst sll header bytes (- (length-field-shift type) 2))
+    ;; The specified EXTRA value is the exact value placed in the header
+    ;; as the word count when allocating code.
+    (cond ((= type code-header-widetag)
+           (inst add header header type))
+          (t
+           (inst add header header (+ (ash -2 (length-field-shift type)) type))
+           (inst and bytes (lognot lowtag-mask))))
     (pseudo-atomic ()
-      (allocation result bytes lowtag :temp-tn temp)
+      (allocation nil bytes lowtag result :temp-tn temp)
       (storew header result 0 lowtag))))
