@@ -43,9 +43,10 @@
 # include <dlfcn.h>
 #endif
 
-static int decode_locs(lispobj packed_integer, int *offset, int *elsewhere)
+int df_decode_locs(lispobj encoded, int *offset, int *elsewhere)
 {
     struct varint_unpacker unpacker;
+    lispobj packed_integer = listp(encoded) ? CONS(encoded)->cdr : encoded;
     varint_unpacker_init(&unpacker, packed_integer);
     return varint_unpack(&unpacker, offset) && varint_unpack(&unpacker, elsewhere);
 }
@@ -67,14 +68,14 @@ debug_function_from_pc (struct code* code, void *pc)
 
     struct compiled_debug_fun *df = (struct compiled_debug_fun*)native_pointer(di->fun_map);
     int begin, end, elsewhere_begin, elsewhere_end;
-    if (!decode_locs(df->encoded_locs, &begin, &elsewhere_begin))
+    if (!df_decode_locs(df->encoded_locs, &begin, &elsewhere_begin))
         return NULL;
     sword_t offset = (char*)pc - code_text_start(code);
     while (df) {
         struct compiled_debug_fun *next;
         if (df->next != NIL) {
             next = (struct compiled_debug_fun*) native_pointer(df->next);
-            if (!decode_locs(next->encoded_locs, &end, &elsewhere_end))
+            if (!df_decode_locs(next->encoded_locs, &end, &elsewhere_end))
                 return NULL;
         } else {
             next = 0;
@@ -121,11 +122,23 @@ lispobj debug_print(lispobj string)
     return 0;
 }
 
-static int string_equal (struct vector *vector, char *string)
+lispobj symbol_package(struct symbol* s)
 {
-    if (widetag_of(&vector->header) != SIMPLE_BASE_STRING_WIDETAG)
-        return 0;
-    return !strcmp((char *) vector->data, string);
+    static int warned;
+    // If using ldb when debugging cold-init, this can be confusing to see all symbols
+    // as if they were uninterned, but package-IDs are always available in the symbol.
+    // End-users should never see this failure.
+    if (!lisp_package_vector) {
+        if (!warned) {
+          fprintf(stderr, "Warning: package vector has not been initialized yet\n");
+          warned = 1;
+        }
+        return NIL;
+    }
+    struct vector* v = VECTOR(lisp_package_vector);
+    int id = symbol_package_id(s);
+    if (id < vector_len(v)) return v->data[id];
+    lose("can't decode package ID %d", id);
 }
 
 static void
@@ -148,26 +161,24 @@ print_entry_name (lispobj name, FILE *f)
         putc(')', f);
     } else if (lowtag_of(name) == OTHER_POINTER_LOWTAG) {
         struct symbol *symbol = SYMBOL(name);
+        char* prefix = 0;
         int widetag = header_widetag(symbol->header);
         switch (widetag) {
         case SYMBOL_WIDETAG:
-            if (symbol->package != NIL) {
-                struct package *pkg
-                    = (struct package *) native_pointer(follow_maybe_fp(symbol->package));
-                struct vector *pkg_name = VECTOR(follow_maybe_fp(pkg->_name));
-                if (string_equal(pkg_name, "COMMON-LISP"))
-                    ;
-                else if (string_equal(pkg_name, "COMMON-LISP-USER")) {
-                    fputs("CL-USER::", f);
-                }
-                else if (string_equal(pkg_name, "KEYWORD")) {
-                    putc(':', f);
-                } else {
-                    print_string(pkg_name, f);
-                    fputs("::", f);
-                }
+            switch (symbol_package_id(symbol)) {
+            case PACKAGE_ID_NONE: prefix = "#:"; break;
+            case PACKAGE_ID_LISP: prefix = ""; break;
+            case PACKAGE_ID_USER: prefix = "CL-USER::"; break;
+            case PACKAGE_ID_KEYWORD: prefix = ":"; break;
             }
-            print_string(VECTOR(follow_maybe_fp(symbol->name)), f);
+            if (prefix) fputs(prefix, f); else {
+                struct package *pkg
+                    = (struct package *)native_pointer(symbol_package(symbol));
+                struct vector *pkg_name = VECTOR(follow_maybe_fp(pkg->_name));
+                print_string(pkg_name, f);
+                fputs("::", f);
+            }
+            print_string(symbol_name(symbol), f);
             break;
         case SIMPLE_BASE_STRING_WIDETAG:
 #ifdef SIMPLE_CHARACTER_STRING_WIDETAG
@@ -260,7 +271,7 @@ call_info_from_context(struct call_info *info, os_context_t *context)
     } else
 #endif
     {
-        pc = *os_context_pc_addr(context);
+        pc = os_context_pc(context);
         info->frame =
             (struct call_frame *)(uword_t)
                 (*os_context_register_addr(context, reg_CFP));
@@ -384,6 +395,13 @@ lisp_backtrace(int nframes)
             && info.lra != NIL)
             printf("LRA=%p ", (void*)info.lra);
 
+        int fpvalid = (lispobj*)info.frame >= thread->control_stack_start
+          && (lispobj*)info.frame < thread->control_stack_end;
+
+        // If the FP is invalid, then quite likely we'd crash trying to find a
+        // compiled-debug-fun because info.code is a wild pointer
+        if (!fpvalid) { printf(" BAD FRAME\n"); break; }
+
         if (info.code) {
             struct compiled_debug_fun *df;
             if (absolute_pc &&
@@ -564,7 +582,7 @@ void backtrace_from_fp(void *fp, int nframes, int start) {
 
 void backtrace_from_context(os_context_t *context, int nframes) {
     void *fp = (void *)os_context_frame_pointer(context);
-    print_backtrace_frame((void *)*os_context_pc_addr(context), fp, 0, stdout);
+    print_backtrace_frame((void *)os_context_pc(context), fp, 0, stdout);
     backtrace_from_fp(fp, nframes - 1, 1);
 }
 
@@ -606,7 +624,7 @@ int simple_fun_index_from_pc(struct code* code, char *pc)
 static boolean __attribute__((unused)) print_lisp_fun_name(char* pc)
 {
   struct code* code;
-  if (gc_managed_addr_p((uword_t)pc) &&
+  if (gc_managed_heap_space_p((uword_t)pc) &&
       (code = (void*)component_ptr_from_pc(pc)) != 0) {
       struct compiled_debug_fun* df = debug_function_from_pc(code, pc);
       if (df) {
